#! /usr/bin/env python3

##
# Class implementing Xcom protocol 
##

import struct
from io import BufferedWriter, BufferedReader, BytesIO

from .parameters import ERROR_CODES

class Service:

    object_type: bytes
    object_id: int
    property_id: bytes
    property_data: bytes

    @staticmethod
    def parse(f: BufferedReader):
        return Service(
            f.read(2),
            readUInt(f),
            f.read(2),
            f.read(-1)
        )

    def __init__(self, 
            object_type: bytes, object_id: int, 
            property_id: bytes, property_data: bytes):

        assert len(object_type) == 2
        assert len(property_id) == 2

        self.object_type = object_type
        self.object_id = object_id
        self.property_id = property_id
        self.property_data = property_data

    def assemble(self, f: BufferedWriter):
        f.write(self.object_type)
        writeUInt(f, self.object_id)
        f.write(self.property_id)
        f.write(self.property_data)

    def __len__(self) -> int:
        return 2*2 + 4 + len(self.property_data)

    def __str__(self) -> str:
        return f"(obj_type={self.object_type}, obj_id={self.object_id}, property_id={self.property_id}, property_data={self.property_data})"

class Frame:

    service_flags: int
    service_id: bytes
    service_data: Service

    @staticmethod
    def parse(f: BufferedReader):
        return Frame(
            service_flags=readUChar(f),
            service_id=f.read(1),
            service_data=Service.parse(f)
        )

    @staticmethod
    def parseBytes(buf: bytes):
        return Frame.parse(BytesIO(buf))

    def __init__(self, service_id: bytes, service_data: Service, service_flags=0):
        assert service_flags >= 0, "service_flag must not be negative"
        assert len(service_id) == 1

        self.service_flags = service_flags
        self.service_id = service_id
        self.service_data = service_data

    def assemble(self, f: BufferedWriter):
        writeUChar(f, self.service_flags)
        f.write(self.service_id)
        self.service_data.assemble(f)

    def getBytes(self) -> bytes:
        buf = BytesIO()
        self.assemble(buf)
        return buf.getvalue()

    def __len__(self) -> int:
        return 2*1 + len(self.service_data)

    def __str__(self) -> str:
        return f"Frame(flags={self.service_flags}, id={self.service_id}, setr={self.service_data})"

class Header:

    frame_flags: int
    src_addr: int
    dst_addr: int
    data_length: int

    length: int = 2*4 + 2 + 1

    @staticmethod
    def parse(f: BufferedReader):
        return Header(
            frame_flags=readUChar(f),
            src_addr=readUInt(f),
            dst_addr=readUInt(f),
            data_length=readUShort(f)
        )

    @staticmethod
    def parseBytes(buf: bytes):
        return Header.parse(BytesIO(buf))

    def __init__(self, src_addr: int, dst_addr: int, data_length: int, frame_flags=0):
        assert frame_flags >= 0, "frame_flags must not be negative"

        self.frame_flags = frame_flags
        self.src_addr = src_addr
        self.dst_addr = dst_addr
        self.data_length = data_length

    def assemble(self, f: BufferedWriter):
        writeUChar(f, self.frame_flags)
        writeUInt(f, self.src_addr)
        writeUInt(f, self.dst_addr)
        writeUShort(f, self.data_length)

    def getBytes(self) -> bytes:
        buf = BytesIO()
        self.assemble(buf)
        return buf.getvalue()

    def __len__(self) -> int:
        return self.length

    def __str__(self) -> str:
        return f"Header(flags={self.frame_flags}, src={self.src_addr}, dst={self.dst_addr}, data_length={self.data_length})"

class Package:

    start_byte: bytes = b'\xAA'
    header: Header
    frame_data: Frame

    @staticmethod
    def parse(f: BufferedReader):
        # package sometimes starts with 0xff
        if (sb := f.read(1)) == b'\xff':
            assert f.read(1) == Package.start_byte
        else:
            assert sb == Package.start_byte

        h_raw = f.read(Header.length)
        assert checksum(h_raw) == f.read(2)
        header = Header.parseBytes(h_raw)

        f_raw = f.read(header.data_length)
        assert checksum(f_raw) == f.read(2)
        frame = Frame.parseBytes(f_raw)

        return Package(header, frame)

    @staticmethod
    def parseBytes(buf: bytes):
        return Package.parse(BytesIO(buf))

    @staticmethod
    def genPackage(service_id: bytes,
            object_id: int,
            object_type: bytes,
            property_id: bytes,
            property_data: bytes,
            src_addr = 1,
            dst_addr = 0):
        
        frame = Frame(
            service_id, 
            Service(object_type, object_id, property_id, property_data)
        )

        return Package(
            Header(src_addr, dst_addr, len(frame)),
            frame
        )


    def __init__(self, header: Header, frame_data: Frame):
        self.header = header
        self.frame_data = frame_data

    def assemble(self, f: BufferedWriter):
        f.write(self.start_byte)

        header = self.header.getBytes()
        f.write(header)
        f.write(checksum(header))

        data = self.frame_data.getBytes()
        f.write(data)
        f.write(checksum(data))

    def getBytes(self) -> bytes:
        buf = BytesIO()
        self.assemble(buf)
        return buf.getvalue()

    def isResponse(self) -> bool:
        return (self.frame_data.service_flags & 2) >> 1 == 1

    def isError(self) -> bool:
        return self.frame_data.service_flags & 1 == 1

    def getError(self) -> str:
        if self.isError():
            return ERROR_CODES.get(
                self.frame_data.service_data.property_data,
                "UNKNOWN ERROR"
            )
        return None
 
    def __str__(self) -> str:
        return f"Package(header={self.header}, frame_data={self.frame_data})"

##

def checksum(data: bytes) -> bytes:
    """Function to calculate the checksum needed for the header and the data"""
    A = 0xFF
    B = 0x00

    for d in data:
        A = (A + d) % 0x100
        B = (B + A) % 0x100

    A = struct.pack("<B", A)
    B = struct.pack("<B", B)

    return A + B

##

def readUInt(f: BufferedReader) -> int:
    return int.from_bytes(f.read(4), byteorder="little", signed=False)

def writeUInt(f: BufferedWriter, value: int) -> int:
    return f.write(value.to_bytes(4, byteorder="little", signed=False))

def readSInt(f: BufferedReader) -> int:
    return int.from_bytes(f.read(4), byteorder="little", signed=True)

def writeSInt(f: BufferedWriter, value: int) -> int:
    return f.write(value.to_bytes(4, byteorder="little", signed=True))


def readUShort(f: BufferedReader) -> int:
    return int.from_bytes(f.read(2), byteorder="little", signed=False)

def writeUShort(f: BufferedWriter, value: int) -> int:
    return f.write(value.to_bytes(2, byteorder="little", signed=False))


def readUChar(f: BufferedReader) -> int:
    return int.from_bytes(f.read(1), byteorder="little", signed=False)

def writeUChar(f: BufferedWriter, value: int) -> int:
    return f.write(value.to_bytes(1, byteorder="little", signed=False))



#! /usr/bin/env python3

import socket
import logging

from concurrent.futures import ThreadPoolExecutor

from .protocol import Package
from .XcomAbs import XcomAbs

##
# Class abstracting Xcom-LAN TCP network protocol
##

class XcomLANTCP(XcomAbs):

    def __init__(self, port=4001):
        """
        MOXA is connecting to the TCP Server we are creating here.

        Once it is connected we can send package requests.
        """

        self.localPort = port
        self.log = logging.getLogger("XcomLANTCP")

    def __enter__(self):
        self.log.info(f"Starting TCP server on port {self.localPort}")

        self.tcpServer = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.tcpServer.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.tcpServer.bind(("", self.localPort))
        self.tcpServer.listen(1)

        self.log.info("Waiting for MOXA to connect...")
        
        conn, addr = self.tcpServer.accept()
        self.log.debug(f"Got connection from {addr}")

        self.conn = conn

        # TODO handshake with GUID (?)
        return self

    def __exit__(self, *_):
        self.conn.close()
        self.tcpServer.close()
        return True

    def sendPackage(self, package: Package) -> Package:
        data: bytes = package.getBytes()
        
        self.log.debug(f" --> {data.hex()}")
        self.conn.send(data)

        response: bytes = self.conn.recv(256)
        self.log.debug(f" <-- {response.hex()}")
        
        retPackage = Package.parseBytes(response)
        self.log.debug(retPackage)

        # MOXA sometimes sends unrelated data, so we need to ignore those
        # and resent our request
        try:
            assert retPackage.isResponse()
            assert retPackage.frame_data.service_id == package.frame_data.service_id
            assert retPackage.frame_data.service_data.object_id == package.frame_data.service_data.object_id
        except AssertionError:
            return self.sendPackage(package)

        if err := retPackage.getError():
            raise KeyError("Error received", err)

        return retPackage


##
# Class abstracting Xcom-LAN UDP network protocol
##

class XcomLANUDP(XcomAbs):

    def __init__(self, serverIP: str, dstPort=4002, srcPort=4001):
        """
        Package requests are being sent to serverIP : dstPort using UDP protocol.

        The srcPort is needed, because XcomLAN will send the UDP response NOT
        to the corresponding UDP source endpoint (like every fucking server on this
        planet would do) but rather to <yourIP> : srcPort.

        So in order to make this work, we need to listen on srcPort for incoming
        data.
        """

        self.serverAddress = (serverIP, dstPort)
        self.clientPort = srcPort
        self.log = logging.getLogger("XcomLAN")

        self.udpListener = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.udpListener.bind(("", self.clientPort))
        self.udpListener.settimeout(2) # as recommended by Studer Xcom documentation

    def sendPackage(self, package: Package) -> Package:
        data: bytes = package.getBytes()
        
        with ThreadPoolExecutor() as listener, \
                socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sender:
            self.log.debug(f" --> {data.hex()}")

            future = listener.submit(self._awaitUDPResponse, self.udpListener)

            sender.sendto(data, self.serverAddress)

            value = future.result(10) 
            self.log.debug(f" <-- {value}")

        if not value:
            raise ValueError("Package listener returned None")

        retPackage = Package.parseBytes(value)
        self.log.debug(retPackage)

        if err := retPackage.getError():
            raise KeyError("Error received", err)
        
        return retPackage

    def _awaitUDPResponse(self, udpSocket: socket.socket) -> bytes:
        try:
            response = udpSocket.recv(256)
        except socket.timeout:
            self.log.error("Waiting for response from XcomLAN timed out")
            raise

        return response



#! /usr/bin/env python3

##
# Abstract base class used by various implementations of the Xcom protocol
##

import logging

from abc import ABC, abstractmethod

from .parameters import *
from .protocol import Package

MSG_MAX_LENGTH = 256 # from Xcom Proto Documentation

class XcomAbs(ABC):

    def __init__(self):
        self.log = logging.getLogger("XcomAbs")

    def getValueByID(self, id: int, type: str, dstAddr=100, propertyID=QSP_UNSAVED_VALUE):
        return self.getValue(Datapoint(id, "", type), dstAddr, propertyID)

    def getValue(self, parameter: Datapoint, dstAddr=100, propertyID=QSP_UNSAVED_VALUE):
        self.log.debug(f"requesting value {parameter}")

        objectType = TYPE_PARAMETER

        if 3000 <= parameter.id <= 3168:
            objectType = TYPE_INFO
        elif parameter.id >= 7000:
            objectType = TYPE_INFO
        
        request: Package = Package.genPackage(
            service_id=PROPERTY_READ,
            object_id=parameter.id,
            object_type=objectType,
            property_id=propertyID,
            property_data=b'',
            dst_addr=dstAddr
        )

        response: Package = self.sendPackage(request)

        return parameter.unpackValue(response.frame_data.service_data.property_data)        

    def setValue(self, parameter: Datapoint, value, dstAddr=100, propertyID=QSP_UNSAVED_VALUE):
        self.log.debug(f"setting value {parameter}")

        request: Package = Package.genPackage(
            service_id=PROPERTY_WRITE,
            object_id=parameter.id,
            object_type=TYPE_PARAMETER,
            property_id=propertyID,
            property_data=parameter.packValue(value),
            dst_addr=dstAddr
        )

        self.sendPackage(request)

    ## TODO
    #def setProperty():
    #    raise NotImplementedError
    

    @abstractmethod
    def sendPackage(self, package: Package)  -> Package:
        raise NotImplementedError


#! /usr/bin/env python3

##
# Class abstracting Xcom-RS232i serial protocol
##

import serial
import logging

from .parameters import ERROR_CODES
from .protocol import Package
from .XcomAbs import XcomAbs

SERIAL_TERMINATOR = b'\x0D\x0A'


class XcomRS232(XcomAbs):

    def __init__(self, serialDevice: str, baudrate: int, timeout=2):
        self.serialDevice = serialDevice
        self.baudrate = baudrate
        self.timeout = timeout
        self.log = logging.getLogger("XcomRS232")

    def sendPackage(self, package: Package) -> Package:
        data: bytes = package.getBytes() + SERIAL_TERMINATOR

        with serial.Serial(self.serialDevice, self.baudrate, timeout=self.timeout) as ser:
            self.log.debug(f" --> {data.hex()}")
            ser.write(data)

            response: bytes = ser.read_until(SERIAL_TERMINATOR, size=256)
            self.log.debug(f" <-- {response.hex()}")

        retPackage = Package.parseBytes(response[:-len(SERIAL_TERMINATOR)])
        self.log.debug(retPackage)

        if err := retPackage.getError():
            raise KeyError("Error received", err)

        return retPackage



#! /usr/bin/env python3

##
# Definition of all parameters / constants used in the Xcom protocol
##

import struct
from dataclasses import dataclass


class UnknownDatapointException(Exception):
    pass

@dataclass
class ValueTuple:
    id: int
    value: str

    def __eq__(self, __o: object) -> bool:
        if __o.__class__ is self.__class__:
            return __o.id == self.id
        return __o == self.id

    def __ne__(self, __o: object) -> bool:
        if __o.__class__ is self.__class__:
            return __o.id != self.id
        return __o != self.id

    def __str__(self) -> str:
        return self.value

@dataclass
class Datapoint:
    id: int
    name: str
    type: str
    unit: str = ""

    def __eq__(self, __o: object) -> bool:
        if __o.__class__ is self.__class__:
            return __o.id == self.id
        return __o == self.id

    def __ne__(self, __o: object) -> bool:
        if __o.__class__ is self.__class__:
            return __o.id != self.id
        return __o != self.id

    def unpackValue(self, value: bytes):
        if self.type is TYPE_FLOAT:
            return struct.unpack("<f", value)[0]
        if self.type is TYPE_SINT:
            return struct.unpack("<i", value)[0]
        if self.type is TYPE_BOOL:
            return struct.unpack("<?", value)[0]
        if self.type is TYPE_SHORT_ENUM:
            return struct.unpack("<h", value)[0]

        raise TypeError("Unknown datatype", self)

    def packValue(self, value) -> bytes:
        if self.type is TYPE_FLOAT:
            return struct.pack("<f", value)
        if self.type is TYPE_SINT:
            return struct.pack("<i", value)
        if self.type is TYPE_BOOL:
            return struct.pack("<?", value)
        if self.type is TYPE_SHORT_ENUM:
            return struct.pack("<H", value)

        raise TypeError("Unknown datatype", self)

    @staticmethod
    def unpackValueByID(id: int, value: bytes):
        dataPoint = Dataset.getParamByID(id)
        return dataPoint.unpackValue(value)


### data types
TYPE_BOOL       = "BOOL"
TYPE_SINT       = "INTEGER"
TYPE_FLOAT      = "FLOAT"
TYPE_SHORT_ENUM = "ENUM_SHORT"
TYPE_LONG_ENUM  = "ENUM_LONG"
TYPE_STRING     = "STRING"
TYPE_BYTES      = "BYTES"

### service_id
PROPERTY_READ   = b'\x01'
PROPERTY_WRITE  = b'\x02'

### object_type
TYPE_INFO       = b'\x01\x00'
TYPE_PARAMETER  = b'\x02\x00'
TYPE_MESSAGE    = b'\x03\x00'
TYPE_GUID       = b'\x04\x00'
TYPE_DATALOG    = b'\x05\x00'

### property_id
QSP_VALUE           = b'\x05\x00'
QSP_MIN             = b'\x06\x00'
QSP_MAX             = b'\x07\x00'
QSP_LEVEL           = b'\x08\x00'
QSP_UNSAVED_VALUE   = b'\x0D\x00'

## values for QSP_LEVEL
QSP_LEVEL_VIEW_ONLY     = b'\x00\x00'
QSP_LEVEL_BASIC         = b'\x10\x00'
QSP_LEVEL_EXPERT        = b'\x20\x00'
QSP_LEVEL_INSTALLER     = b'\x30\x00'
QSP_LEVEL_QSP           = b'\x40\x00'


### operating modes (11016)
MODE_NIGHT      = ValueTuple(0, "MODE_NIGHT")
MODE_STARTUP    = ValueTuple(1, "MODE_STARTUP")
MODE_CHARGER    = ValueTuple(3, "MODE_CHARGER")
MODE_SECURITY   = ValueTuple(5, "MODE_SECURITY")
MODE_OFF        = ValueTuple(6, "MODE_OFF")
MODE_CHARGE     = ValueTuple(8, "MODE_CHARGE")
MODE_CHARGE_V   = ValueTuple(9, "MODE_CHARGE_V")
MODE_CHARGE_I   = ValueTuple(10, "MODE_CHARGE_I")
MODE_CHARGE_T   = ValueTuple(11, "MODE_CHARGE_T")

MODE_CHARGING = (
    MODE_CHARGE,
    MODE_CHARGE_V,
    MODE_CHARGE_I,
    MODE_CHARGE_T
)

### battey cycle phase (11038)
PHASE_BULK      = ValueTuple(0, "PHASE_BULK")
PHASE_ABSORPT   = ValueTuple(1, "PHASE_ABSORPT")
PHASE_EQUALIZE  = ValueTuple(2, "PHASE_EQUALIZE")
PHASE_FLOATING  = ValueTuple(3, "PHASE_FLOATING")
PHASE_R_FLOAT   = ValueTuple(6, "PHASE_R_FLOAT")
PHASE_PER_ABS   = ValueTuple(7, "PHASE_PER_ABS")


### error codes
ERROR_CODES = {
    b'\x01\x00': "INVALID_FRAME",
    b'\x02\x00': "DEVICE_NOT_FOUND",
    b'\x03\x00': "RESPONSE_TIMEOUT",
    b'\x11\x00': "SERVICE_NOT_SUPPORTED",
    b'\x12\x00': "INVALID_SERVICE_ARGUMENT",
    b'\x13\x00': "SCOM_ERROR_GATEWAY_BUSY",
    b'\x21\x00': "TYPE_NOT_SUPPORTED",
    b'\x22\x00': "OBJECT_ID_NOT_FOUND",
    b'\x23\x00': "PROPERTY_NOT_SUPPORTED",
    b'\x24\x00': "INVALID_DATA_LENGTH",
    b'\x25\x00': "PROPERTY_IS_READ_ONLY",
    b'\x26\x00': "INVALID_DATA",
    b'\x27\x00': "DATA_TOO_SMALL",
    b'\x28\x00': "DATA_TOO_BIG",
    b'\x29\x00': "WRITE_PROPERTY_FAILED",
    b'\x2A\x00': "READ_PROPERTY_FAILED",
    b'\x2B\x00': "ACCESS_DENIED",
    b'\x2C\x00': "SCOM_ERROR_OBJECT_NOT_SUPPORTED",
    b'\x2D\x00': "SCOM_ERROR_MULTICAST_READ_NOT_SUPPORTED",
    b'\x2E\x00': "OBJECT_PROPERTY_INVALID",
    b'\x2F\x00': "FILE_OR_DIR_NOT_PRESENT",
    b'\x30\x00': "FILE_CORRUPTED",
    b'\x81\x00': "INVALID_SHELL_ARG",
}

### main parameters

class Dataset:

    SMART_BOOST_ALLOWED = Datapoint(1126, "SMART_BOOST_ALLOWED", TYPE_BOOL)
    BATTERY_CHARGE_CURR = Datapoint(1138, "BATTERY_CHARGE_CURR", TYPE_FLOAT)
    SMART_BOOST_LIMIT = Datapoint(1607, "SMART_BOOST_LIMIT", TYPE_FLOAT)

    PARAMS_SAVED_IN_FLASH = Datapoint(1550, "PARAMS_SAVED_IN_FLASH", TYPE_BOOL)

    USER_LEVEL = Datapoint(5012, "USER_LEVEL", TYPE_SHORT_ENUM)

    AC_ENERGY_IN_CURR_DAY = Datapoint(3081, "AC_POWER_IN_CURR_DAY", TYPE_FLOAT, "kWh")
    AC_ENERGY_IN_PREV_DAY = Datapoint(3080, "AC_POWER_IN_PREV_DAY", TYPE_FLOAT, "kWh")
    AC_ENERGY_OUT_CURR_DAY = Datapoint(3083, "AC_ENERGY_OUT_CURR_DAY", TYPE_FLOAT, "kWh")
    AC_ENERGY_OUT_PREV_DAY = Datapoint(3082, "AC_ENERGY_OUT_PREV_DAY", TYPE_FLOAT, "kWh")
    AC_FREQ_IN = Datapoint(3084, "AC_FREQ_IN", TYPE_FLOAT, "Hz")
    AC_FREQ_OUT = Datapoint(3085, "AC_FREQ_OUT", TYPE_FLOAT, "Hz")
    AC_POWER_IN = Datapoint(3137, "AC_POWER_IN", TYPE_FLOAT, "kW")
    AC_POWER_OUT = Datapoint(3136, "AC_POWER_OUT", TYPE_FLOAT, "kW")
    AC_VOLTAGE_IN = Datapoint(3011, "AC_VOLTAGE_IN", TYPE_FLOAT, "V")
    AC_VOLTAGE_OUT = Datapoint(3021, "AC_VOLTAGE_OUT", TYPE_FLOAT, "V")

    BATT_VOLTAGE = Datapoint(7000, "BATT_VOLTAGE", TYPE_FLOAT, "V")
    BATT_CURRENT = Datapoint(7001, "BATT_CURRENT", TYPE_FLOAT, "A")
    BATT_SOC = Datapoint(7032, "BATT_SOC", TYPE_FLOAT, "%")
    BATT_TEMP = Datapoint(7029, "BATT_TEMP", TYPE_FLOAT, "°C")
    BATT_CYCLE_PHASE = Datapoint(11038, "BATT_CYCLE_PHASE", TYPE_SHORT_ENUM)

    PV_VOLTAGE = Datapoint(11041 ,"PV_VOLTAGE", TYPE_FLOAT, "V")
    PV_POWER = Datapoint(11043, "PV_POWER", TYPE_FLOAT, "W")
    PV_ENERGY_CURR_DAY = Datapoint(11007, "PV_ENERGY_CURR_DAY", TYPE_FLOAT, "kWh")
    PV_ENERGY_PREV_DAY = Datapoint(11011, "PV_ENERGY_CURR_DAY", TYPE_FLOAT, "kWh")
    PV_ENERGY_TOTAL = Datapoint(11009, "PV_ENERGY_TOTAL", TYPE_FLOAT, "MWh")
    PV_SUN_HOURS_CURR_DAY = Datapoint(11025, "PV_SUN_HOURS_CURR_DAY", TYPE_FLOAT, "h")
    PV_SUN_HOURS_PREV_DAY = Datapoint(11026, "PV_SUN_HOURS_PREV_DAY", TYPE_FLOAT, "h")

    PV_OPERATION_MODE = Datapoint(11016, "PV_OPERATION_MODE", TYPE_SHORT_ENUM)
    PV_NEXT_EQUAL = Datapoint(11037, "PV_NEXT_EQUAL", TYPE_FLOAT, "d")

    FORCE_NEW_CYCLE = Datapoint(10029, "FORCE_NEW_CYCLE", TYPE_SINT)

    @staticmethod
    def getParamByID(id: int) -> Datapoint:
        for point in Dataset._getDatapoints():
            if point == id:
                return point

        raise UnknownDatapointException(id)

    @staticmethod
    def _getDatapoints() -> list[Datapoint]:
        points = list()
        for val in Dataset.__dict__.values():
            if type(val) is Datapoint:
                points.append(val)

        return points
